use std::borrow::ToOwned;
use std::sync::Arc;

use log::warn;
use once_cell::sync::Lazy;
use parse_display::Display;
use regex::Regex;
use serde::Deserialize;
use tokio::select;
use tokio::sync::broadcast;

use deckster_shared::handler_communication::HandlerCommand;
use deckster_shared::handler_communication::{KnobEvent, RotationDirection};
use deckster_shared::path::KnobPath;
use deckster_shared::state::KnobStyleByStateMap;
use pa_volume_interface::{PaEntityKind, PaEntityMetadata, PaEntityState, PaVolumeInterface};

#[derive(Debug, Deserialize)]
pub struct Config {
    pub target: Target,
    pub delta: Option<f32>,
    #[serde(default)]
    pub disable_press_to_mute: bool,
    #[serde(default)]
    pub disable_press_to_unmute: bool,
    #[serde(default)]
    pub muted_turn_action: MutedTurnAction,
    #[serde(default)]
    pub style: KnobStyleByStateMap<State>,
}

#[derive(Debug, Eq, PartialEq, Deserialize, Display)]
#[serde(rename_all = "kebab-case")]
#[display(style = "kebab-case")]
pub enum TargetKind {
    Input,
    Output,
    Application,
}

#[derive(Debug, Deserialize)]
pub struct Target {
    #[serde(rename = "type")]
    kind: TargetKind,
    predicates: Vec<TargetPredicate>,
}

#[derive(Debug, Deserialize)]
pub struct TargetPredicate {
    property: TargetPredicateProperty,
    #[serde(flatten)]
    pattern: TargetPredicatePattern,
}

#[derive(Debug, Deserialize)]
#[serde(untagged)]
pub enum TargetPredicatePattern {
    Static {
        value: String,
    },
    Regex {
        #[serde(with = "serde_regex")]
        regex: Regex,
    },
}

#[derive(Debug, Eq, PartialEq, Deserialize, Display)]
#[serde(rename_all = "kebab-case")]
#[display(style = "kebab-case")]
pub enum TargetPredicateProperty {
    Description,
    InternalName,
    ApplicationName,
    BinaryName,
}

#[derive(Debug, Default, Eq, PartialEq, Deserialize)]
#[serde(rename_all = "kebab-case")]
pub enum MutedTurnAction {
    #[default]
    Ignore,
    Normal,
    UnmuteAtZero,
    Unmute,
}

#[derive(Debug, Default, Eq, PartialEq, Deserialize)]
#[serde(rename_all = "kebab-case")]
pub enum Direction {
    #[default]
    Output,
    Input,
}

#[derive(Debug, Eq, PartialEq, Hash, Deserialize)]
#[serde(rename_all = "kebab-case")]
pub enum State {
    Inactive,
    Active,
    Muted,
}

static PA_VOLUME_INTERFACE: Lazy<PaVolumeInterface> = Lazy::new(|| PaVolumeInterface::spawn_thread("deckster".to_owned()));

fn get_volume_from_cv(channel_volumes: &[f32]) -> f32 {
    *channel_volumes.iter().max_by(|a, b| a.partial_cmp(b).unwrap()).unwrap()
}

fn state_matches(target: &Target, state: &PaEntityState) -> bool {
    if !match target.kind {
        TargetKind::Input => state.kind() == PaEntityKind::Source,
        TargetKind::Output => state.kind() == PaEntityKind::Sink,
        TargetKind::Application => state.kind() == PaEntityKind::SinkInput,
    } {
        return false;
    }

    static EMPTY_STRING: String = String::new();

    return target.predicates.iter().all(|p| {
        let v = match (&p.property, state.metadata()) {
            (TargetPredicateProperty::InternalName, PaEntityMetadata::Sink { name, .. }) => Some(name),
            (TargetPredicateProperty::InternalName, PaEntityMetadata::Source { name, .. }) => Some(name),
            (TargetPredicateProperty::InternalName, PaEntityMetadata::SinkInput { .. }) => None,
            (TargetPredicateProperty::Description, PaEntityMetadata::Sink { description, .. }) => Some(description),
            (TargetPredicateProperty::Description, PaEntityMetadata::Source { description, .. }) => Some(description),
            (TargetPredicateProperty::Description, PaEntityMetadata::SinkInput { description, .. }) => Some(description),
            (TargetPredicateProperty::ApplicationName, PaEntityMetadata::Sink { .. }) => None,
            (TargetPredicateProperty::ApplicationName, PaEntityMetadata::Source { .. }) => None,
            (TargetPredicateProperty::ApplicationName, PaEntityMetadata::SinkInput { application_name, .. }) => {
                Some(application_name.as_ref().unwrap_or(&EMPTY_STRING))
            }
            (TargetPredicateProperty::BinaryName, PaEntityMetadata::Sink { .. }) => None,
            (TargetPredicateProperty::BinaryName, PaEntityMetadata::Source { .. }) => None,
            (TargetPredicateProperty::BinaryName, PaEntityMetadata::SinkInput { binary_name, .. }) => Some(binary_name.as_ref().unwrap_or(&EMPTY_STRING)),
        };

        if let Some(v) = v {
            match &p.pattern {
                TargetPredicatePattern::Static { value } => value == v,
                TargetPredicatePattern::Regex { regex } => regex.is_match(v),
            }
        } else {
            warn!("Property \"{}\" is not available for targets of type \"{}\"", &p.property, &target.kind);

            false
        }
    });
}

pub async fn handle(path: KnobPath, config: Arc<Config>, mut events: broadcast::Receiver<(KnobPath, KnobEvent)>, commands: flume::Sender<HandlerCommand>) {
    let mut entity_state: Option<Arc<PaEntityState>> = None;

    let pa_volume_interface = &PA_VOLUME_INTERFACE;
    let (initial_state, mut volume_states) = pa_volume_interface.subscribe_to_state();

    let update_knob_value = {
        let commands = commands.clone();
        let config = Arc::clone(&config);
        let path = path.clone();

        move |entity_state: &Option<Arc<PaEntityState>>| {
            commands
                .send(HandlerCommand::SetKnobValue {
                    path: path.clone(),
                    value: entity_state.as_ref().map(|s| {
                        if s.is_muted() && config.muted_turn_action == MutedTurnAction::UnmuteAtZero {
                            0.0
                        } else {
                            get_volume_from_cv(&s.channel_volumes())
                        }
                    }),
                })
                .unwrap();
        }
    };

    let update_knob_style = {
        let commands = commands.clone();
        let config = Arc::clone(&config);
        let path = path.clone();

        move |entity_state: &Option<Arc<PaEntityState>>| {
            let state = match entity_state {
                None => State::Inactive,
                Some(s) if s.is_muted() => State::Muted,
                Some(_) => State::Active,
            };

            let mut style = config.style.get(&state).cloned();

            if let Some(ref mut s) = &mut style {
                let v = entity_state.as_ref().map(|s| get_volume_from_cv(&s.channel_volumes()));

                if let Some(ref mut label) = &mut s.label {
                    if let Some(v) = v {
                        // v is only None when state is State::Inactive
                        *label = label.replace("{percentage}", &((v * 100.0).round() as u32).to_string());
                    }
                }
            }

            commands
                .send(HandlerCommand::SetKnobStyle {
                    path: path.clone(),
                    value: style,
                })
                .unwrap();
        }
    };

    if let Some(state) = initial_state {
        entity_state = state
            .entities_by_id()
            .values()
            .find(|entity| state_matches(&config.target, &entity))
            .map(Arc::clone);
    }

    loop {
        select! {
            Ok(volume_state) = volume_states.recv() => {
                entity_state = volume_state.entities_by_id().values().find(|entity| state_matches(&config.target, &entity)).map(Arc::clone);
                update_knob_style(&entity_state);
                update_knob_value(&entity_state);
            }

            Ok((event_path, event)) = events.recv() => {
                if event_path != path {
                    continue;
                }

                if let Some(entity_state) = &entity_state {
                    match event {
                        KnobEvent::Rotate { direction } => {
                            let factor: f32 = match direction {
                                RotationDirection::Clockwise => 1.0,
                                RotationDirection::Counterclockwise => -1.0,
                            };

                            let mut current_v = get_volume_from_cv(&entity_state.channel_volumes());

                            if entity_state.is_muted() {
                                match config.muted_turn_action {
                                    MutedTurnAction::Ignore => continue,
                                    MutedTurnAction::UnmuteAtZero => {
                                        current_v = 0.0
                                    }
                                    MutedTurnAction::Unmute => {},
                                    MutedTurnAction::Normal => {}
                                }
                            }

                            let new_v = (current_v + (factor * config.delta.unwrap_or(0.01))).clamp(0.0, 1.0);
                            if new_v > 0.0 && matches!(config.muted_turn_action, MutedTurnAction::Unmute | MutedTurnAction::UnmuteAtZero) {
                                pa_volume_interface.set_is_muted(*entity_state.id(), false);
                            }

                            pa_volume_interface.set_channel_volumes(*entity_state.id(), vec![new_v; entity_state.channel_volumes().len()]);
                        }
                        KnobEvent::Press => {
                            let is_muted = entity_state.is_muted();

                            if (is_muted && !config.disable_press_to_unmute) || (!is_muted && !config.disable_press_to_mute) {
                                pa_volume_interface.set_is_muted(*entity_state.id(), !is_muted)
                            }
                        }
                        _ => {}
                    }
                }
            }
        }
    }
}
