use std::io;

use either::Either;
use thiserror::Error;

use deckster_shared::handler_communication::HandlerEvent;
use deckster_shared::path::{KeyPath, KnobPath};

#[derive(Debug, Error)]
pub enum RunError {
    #[error("A stdin line could not be read.")]
    LineIo(#[from] io::Error),

    #[error("A stdin line could not be deserialized.")]
    LineDeserialization { line: String, source: toml::de::Error },
}

pub trait DecksterHandler {
    fn handle(&mut self, event: HandlerEvent);
}

pub fn run<H: DecksterHandler, I: FnOnce(im::HashMap<KeyPath, (String, toml::Value)>, im::HashMap<KnobPath, (String, toml::Value)>) -> H>(
    init_handler: I,
) -> Result<(), RunError> {
    let mut handler: Either<H, I> = Either::Right(init_handler);

    for line in io::stdin().lines() {
        let line = line?;

        match handler {
            Either::Left(mut h) => {
                let event: HandlerEvent = toml::from_str(&line).map_err(|e| RunError::LineDeserialization { line, source: e.clone() })?;

                h.handle(event);
                handler = Either::Left(h);
            }
            Either::Right(init_handler) => {
                handler = Either::Left(init_handler(im::HashMap::new(), im::HashMap::new()));
            }
        }
    }

    Ok(())
}
