use std::fmt::{Display, Formatter};
use std::str::FromStr;

use enum_map::Enum;
use enum_ordinalize::Ordinalize;
use serde::{Deserialize, Serialize};
use serde_with::{DeserializeFromStr, SerializeDisplay};
use thiserror::Error;

#[derive(Debug, Eq, PartialEq, Hash, Clone, Copy, SerializeDisplay, DeserializeFromStr)]
/// One-based coordinates of a specific virtual (not physical) key.
pub struct KeyPosition {
    pub x: u16,
    pub y: u16,
}

#[derive(Debug, Error)]
#[error("The input value does not match the required format of <x> and <y> separated by an 'x'")]
pub struct KeyPositionFromStrError {}

impl FromStr for KeyPosition {
    type Err = KeyPositionFromStrError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let values = s.split_once('x');

        if let Some((x, y)) = values {
            if let Ok(x) = u16::from_str(x) {
                if let Ok(y) = u16::from_str(y) {
                    return Ok(KeyPosition { x, y });
                }
            }
        }

        Err(KeyPositionFromStrError {})
    }
}

impl Display for KeyPosition {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        f.write_fmt(format_args!("{}x{}", self.x, self.y))
    }
}

#[derive(Debug, Eq, PartialEq, Hash, Clone, Serialize, Deserialize)]
pub struct KeyPath {
    pub page_id: String,
    pub position: KeyPosition,
}

impl Display for KeyPath {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        f.write_fmt(format_args!("{}/{}", &self.page_id, &self.position))
    }
}

#[derive(Debug, Eq, PartialEq, Hash, Clone, Copy, Serialize, Deserialize, Enum, Ordinalize)]
#[serde(rename_all = "kebab-case")]
pub enum KnobPosition {
    LeftTop,
    LeftMiddle,
    LeftBottom,
    RightTop,
    RightMiddle,
    RightBottom,
}

impl KnobPosition {
    pub fn is_left(&self) -> bool {
        matches!(self, KnobPosition::LeftBottom | KnobPosition::LeftMiddle | KnobPosition::LeftTop)
    }
}

#[derive(Debug, Eq, PartialEq, Hash, Clone, Serialize, Deserialize)]
pub struct KnobPath {
    pub page_id: String,
    pub position: KnobPosition,
}
